#![no_std]
#![no_main]

#![feature(naked_functions)]
//#![feature(asm)]
#![feature(alloc_error_handler)]
#![feature(fn_align)]

//unstable by mzm
#![feature(asm_const)]


// basic configurations 
mod config;
// SBI heap
mod heap;
// hardware abstraction layer 
#[macro_use]
mod hal;
// trap vector
mod trap;

extern crate alloc;

use core::panic::PanicInfo;
#[panic_handler]
#[allow(dead_code)]
fn panic(info: &PanicInfo) ->! {
	println!("\x1b[31;1m[panic]\x1b[0m: {}", info);
	println!("\x1b[31;1m[panic]\x1b[0m: PsicaSBI stops");
	loop {}
}

use config::*;

#[naked]
#[no_mangle]
#[link_section = ".text.init"]
unsafe extern "C" fn _entry() ->! {
	use core::arch::asm;
	asm!(r"
		li sp, {stack_top}
		csrr a0, mhartid
		slli a0, a0, {offset}
		sub sp, sp, a0
		csrw mscratch, sp
		
		csrr a0, mhartid
		call rust_main

		1:
			j 1b
	", 
		stack_top = const KERNEL_ENTRY, 
		offset = const STACK_OFFSET, 
		options(noreturn), 
	)
}

use riscv::register::misa;

#[no_mangle]
#[link_section = ".text.init"]
extern "C" fn rust_main(hartid: usize) {
	if 0 == hartid {
		// init bss section 
		extern "C" {
			static mut _sbss: u32;
			static mut _ebss: u32;
		}
		unsafe {
			r0::zero_bss(&mut _sbss, &mut _ebss);
		}
		heap::init();


		// Initialize PMPCFG0 to PMPCFG3 registers with null values
		// Initialize PMPADDR0 to PMPADDR15 registers with null values
		#[cfg(feature = "k210")]
		{
			hal::sysctl::init();
			hal::sysctl::set_freq();
			hal::fpioa::init();
		}
		hal::uart::init();

		extern "C" {
			fn ekernel();
		}
		let ekernel = ekernel as usize;
		println!("ekernel: {:#x}", ekernel);
		println!("heap_start: {:#x}", config::HEAP_START);
		println!("heap_size: {:#x}", config::HEAP_SIZE);
		assert!(config::HEAP_START > ekernel);

		trap::init();			// install trap handler, this should run per-hart
		println!("trap init");

		println!("{}", LOGO);
		println!(
			"[\x1b[32;1mPsicaSBI\x1b[0m]: Version {}.{}", 
			*SBI_IMPL_VER_MAJOR, 
			*SBI_IMPL_VER_MINOR
		);

		let mideleg: usize;
		let medeleg: usize;
		unsafe {
			use core::arch::asm;
			asm!(
				"csrr {0}, mideleg", 
				out(reg) mideleg
			);
			asm!(
				"csrr {0}, medeleg", 
				out(reg) medeleg
			);
		}
		println!("mideleg: {:#x}, medeleg: {:#x}", mideleg, medeleg);

		let misa = misa::read().unwrap();
		match misa.mxl() {
			misa::MXL::XLEN32 => {
				panic!("RV32 not supported");
			}, 
			misa::MXL::XLEN64 => {
				print!("Extension: RV64");
			}, 
			misa::MXL::XLEN128 => {
				panic!("RV128 tql, not supported yet");
			}
		}
		for ext in 'A'..='Z' {
			if misa.has_extension(ext) {
				print!("{}", ext);
			}
		}
		println!("");
	}
	else {
		trap::init();
		loop {
			unsafe {
				// Because during the boot sequence MIE is not set, 
				// a trap will cause the PC reg to skip wfi instruction, 
				// but SBI won't run into the trap handler. 
				riscv::asm::wfi();
				if riscv::register::mip::read().msoft() {
					hal::clint::clear_ipi(hartid);
					break;
				}
			}
		}
	}

	println!("PsicaSBI hartid {} ready", hartid);

	println!("PMP initialization started.");
const PMP_NAPOT: usize = 0x18; // NAPOT 地址模式
const PMP_R: usize     = 0x04; // 读权限
const PMP_W: usize     = 0x02; // 写权限
const PMP_X: usize     = 0x01; // 执行权限

	unsafe {
		// 定义 PMP 权限配置，包括 NAPOT 模式和读、写、执行权限
		let pmpc: usize = PMP_NAPOT | PMP_R | PMP_W | PMP_X;
	
		// 使用内联汇编，设置临时的异常向量以忽略可能的非法指令陷阱
		// 配置 PMP 来允许对所有内存的访问
		// 恢复原始的异常向量
		use core::arch::asm;
		asm!(
			"la t0, 1f",
			"csrrw t0, mtvec, t0",
			"csrw pmpaddr0, {pmpaddr}",
			"csrw pmpcfg0, {pmpcfg}",
			".align 2",
			"1: csrw mtvec, t0",
			pmpcfg = in(reg) pmpc,
			pmpaddr = in(reg) (-1isize as usize),
			out("t0") _,
		);
		//参考 https://github.com/mit-pdos/xv6-riscv/issues/103
		println!("PMP initialization completed.");
	}
	

	// jump to S-mode kernel
	trap::enter_supervisor(hartid);
}